# -*- coding: utf-8 -*-

"""
Objective of the script: Create a gps track with FloeNavi coordinates from a gps track file.

Usage:

    python floenavi_gpstrack2xy.py <file_path> | <file_search_pattern>

"""

# --- Imports ---

import re
import sys
import glob

from loguru import logger

from floenavi import FloeNavi
from floenavi.dataset import GPSTrack


# --- Globals ---

USAGE_STR = "python floenavi_gpstrack2xy.py <file_path> | <file_search_pattern>"


def floenavi_gpstrack2xy():
    """
    Workflow script for computing FloeNavi (x,y) coordinates based on a a FloeNavi standard GPS track csv file.
    Output will be written in file with additional floenavi tag.
    :return:
    """

    # Get the command line argument
    try:
        file_search = sys.argv[1]
    except IndexError:
        logger.error("Error: missing command line argument -> Terminating")
        logger.info("Usage: %s" % USAGE_STR)
        sys.exit(1)

    # Get the list of files
    file_paths = sorted(glob.glob(file_search))

    # Check number of files to process
    if len(file_paths) == 0:
        logger.error("Error: No such file(s) [%s] -> Terminating" % str(file_search))
        sys.exit(1)
    logger.info("Found %d file(s)" % len(file_paths))

    # Process files
    floenavi_app = FloeNavi()

    # Loop over all files
    counter = 0
    for file_path in file_paths:

        logger.info("Converting: %s" % file_path)

        # Check if file has the correct naming
        if not re.search(r"track\.csv$", file_path):
            logger.error("File %s does not meet naming requirements (*track.csv) -> skipping")
            continue

        # Read the input file
        track = GPSTrack.from_csv(file_path)

        # Compute coordinate transformation
        floenavi_app.get_xy_coordinates(track)

        # Write the output file
        output_filepath = file_path.replace(".csv", "-floenavi.csv")
        track.export_csv(output_filepath)

        # Count successfully converted files
        counter += 1

    # Report and close
    logger.info("Converted %d files" % counter)


if __name__ == "__main__":
    floenavi_gpstrack2xy()
