# -*- coding: utf-8 -*-

import os
import sys
import glob
import yaml
import parse
import pyproj
import numpy as np
from datetime import datetime, timedelta

from floenavi import dataset, polarstern
from loguru import logger

from floenavi.ais import DshipAISExtract

# The algorithm for sea ice coordinate system is implemented in the icedrift module
from icedrift import GeoReferenceStation, GeoPositionData, IceCoordinateSystem


# Get version from VERSION in package root
PACKAGE_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
version_file = open(os.path.abspath(os.path.join(PACKAGE_ROOT_DIR, "VERSION")))
try:
    with version_file as f:
        version = f.read().strip()
except IOError:
    sys.exit("Cannot find VERSION file in package (expected: %s" % version_file)

# Get the location of the config dir
config_dir_def = os.path.abspath(os.path.join(PACKAGE_ROOT_DIR, "CFGDIR"))
try:
    with open(config_dir_def, "r") as f:
        CFG_DIR = f.read().strip()
except IOError:
    print("Cannot find CFGDIR file in package (expected: `%s`)" % config_dir_def)
    CFG_DIR = input("Enter local directory to floenavi config files: ")
    with open(config_dir_def, "w") as f:
        f.write(CFG_DIR)
    print("Written to CFGDIR")

# Get default config files
DEFAULT_CONFIG_FILE = os.path.abspath(os.path.join(CFG_DIR, "floenavi_cfg.yaml"))
DEFAULT_PATHDEF_FILE = os.path.abspath(os.path.join(CFG_DIR, "floenavi_local_path_def.yaml"))

# Package Metadata

__all__ = ["ais", "dataset", "polarstern",   # modules
           "FloeNavi", "FloeNaviConfig",     # main classes
           PACKAGE_ROOT_DIR]

__version__ = version
__author__ = "Stefan Hendricks"
__author_email__ = "stefan.hendricks@awi.de"


class FloeNavi(object):
    """
    The main FloeNavi class and interface for the ice coordinate transformation for floenavi.datasets instances

    Usage:

        floenavi = FloeNavi()

        dataset = dataset.Waypoints(...)
    or  dataset = dataset.GPSTrack(...)

        floenavi.get_xy_coordinates(dataset)

        dataset.export_csv(filename)
    """

    def __init__(self, cfg=None):
        """
        Main FloeNavi app with optional FloeNavi configuration instance.
        :param cfg: (Optional) FloeNaviConfig instance that overrides the default FloeNavi Configuration
        """

        # --- Store arguments ---

        # Use default FloeNavi configuration if no specific configuration is specified
        if cfg is None:
            cfg = FloeNaviConfig()
        self.cfg = cfg

    def get_xy_coordinates(self, dset):
        """
        Compute the (x, y) of the ice coordinate system for a given data set. The data set will be changed in-place
        :param dset: a dataset.WayPoints or dataset.GPSTrack instance
        :return:
        """

        # Check temporal coverage of data set
        msg = "Dataset time range: [%s - %s]" % tuple([time.isoformat() for time in dset.time_bounds])
        logger.info(msg)

        # Temporal coverage of base station data
        # These depend on the temporal coverage of the data set
        floenavi_cfg = self.cfg.get_floenavi_cfg(dset.time_bounds)

        if floenavi_cfg is None:
            logger.error("No FloeNavi configuration found for selected time range in config file -> terminating")
            sys.exit(1)

        logger.info("Requested AIS reference stations: %s" % (", ".join([str(mmsi) for mmsi in floenavi_cfg.mmsis])))

        # Get the (icedrift) reference station data for the selected base stations and data period
        refstat = self.get_virtual_reference_stations(dset.time_bounds, floenavi_cfg)

        # --- Init the ice coordinate system transformation ---
        icecs_trns = IceCoordinateSystem(refstat, **floenavi_cfg.icecs_trns_kwargs)
        logger.info("Initialized Ice Coordinate System (icecs) transformation")
        logger.debug("Settings: %s" % str(floenavi_cfg.icecs_trns_kwargs))

        # --- Convert the coordinates into ice coordinates ---
        pos_data = GeoPositionData(dset.time, dset.longitude, dset.latitude)
        ice_pos, icecs = icecs_trns.get_xy_coordinates(pos_data)
        logger.info("Computed (x,y) coordinates")
        logger.debug("projection parameters: %s" % icecs.proj4str)

        # --- Update the data set in place ----
        dset.update_floenavi_coords(ice_pos.xc, ice_pos.yc)

    def get_virtual_reference_stations(self, dataset_time_bounds, basestation_setup=None,
                                       return_grid_solution=False, **refstat_kwargs):
        """
        Get two virtual reference stations (origin ref and heading ref) from merging
        the information of all available FloeNavi reference stations. The step is
        necessary as the convention of 1 basestation at grid origin and 1 basestation
        at a distance defining a axis of the grid could not be maintained during
        MOSAiC (had to be moved where there was power available). Since the icedrift
        package requires an origin and heading reference station, this two are computed
        as virtual basestation from the actual FloeNavi basestations that have a known
        position on the the FloeNavi grid.

        :param dataset_time_bounds: (datetime list) time range of the dataset
        :param basestation_setup: (FloeNaviBaseStationSetup instance, optional)
        :param return_grid_solution: (bool) If set to true, the method will return the refstation
            and the grid solution
        :return: icedrift.GeoReferenceStation (+ FloeNaviGrid if return_grid_solution is True)
        """

        # Use default base station setup if not specifically supplied
        if basestation_setup is None:
            basestation_setup = self.cfg.get_floenavi_cfg(dataset_time_bounds)

        # Check if base station setup exists
        if basestation_setup is None:
            refstat_ais, fngrid = None, None

        else:
            # Load the collection of FloeNavi basestation files
            logger.info("--- Establish FloeNavi grid ---")
            floenavi_grid = FloeNaviGrid(self.cfg, basestation_setup)
            floenavi_grid.set_time_bounds(dataset_time_bounds)

            # Get Reference station data from two positions
            pos_ref, pos_hdg, fngrid = floenavi_grid.get_virtual_reference_station_positions(
                dataset_time_bounds, **refstat_kwargs)
            if pos_ref is not None:
                refstat_ais = GeoReferenceStation.from_two_basestations(pos_ref, pos_hdg)
                refstat_ais.set_uncertainty(fngrid.origin_uncertainty_m, fngrid.heading_uncertainty_deg)
            else:
                refstat_ais, fngrid = None, None

        # Return value
        if not return_grid_solution:
            return refstat_ais
        else:
            return refstat_ais, fngrid


class FloeNaviConfig(object):
    """
    The configuration of the FloeNavi post processing system, including local paths
    to data and the list of FloeNavi setups (list of basestations and their reference
    positions)
    """

    def __init__(self, yaml_config_file=None, local_path_file=None):
        """
        Loads the FloeNavi configuration (base station setup and local path)
        :param yaml_config_file: (filepath): Optional: link to floenavi config file (base station lists,
            floe navi setups, ...). Only required if a special config file needs to be used. If None,
            the module will use the default config file in the root dir of the floenavi package
        :param local_path_file: (filepath): Optional: link to floenavi input data folder. If None,
            the module will use the default config file in the root dir of the floenavi package
        """

        # --- Save and digest arguments ---

        # Config file with list of base stations and the setups for different periods
        if yaml_config_file is None:
            self.yaml_config_file = DEFAULT_CONFIG_FILE
        else:
            self.yaml_config_file = yaml_config_file
            logger.info("Custom floenavi config file: %s" % self.yaml_config_file)
        if not os.path.isfile(self.yaml_config_file):
            logger.error("Invalid floenavi config file: %s -> terminating" % str(self.yaml_config_file))
            sys.exit(1)

        if local_path_file is None:
            self.local_path_file = DEFAULT_PATHDEF_FILE
        else:
            self.local_path_file = local_path_file
        if not os.path.isfile(self.local_path_file):
            logger.error("Invalid floenavi path definition file: %s -> terminating" % str(self.local_path_file))
            sys.exit(1)

        # Init properties
        self._ref_ctlg = None

        # Init methods
        self._parse_yaml()
        self._create_valid_mmsi_catalog()
        self._create_reference_data_catalog()
        self._create_period_catalog()

    def get_station_ids(self, dataset_time_bounds):
        """
        Searches for the mmsi number of the reference and heading station
        :param dataset_time_bounds: (datetime list) The start and end time of the dataset
        :return:
        """

        # Check time coverage for each FloeNavi configuration period
        # (period section in the floenavi yaml config file)
        mmsis = []
        for period_cfg in self.period_ctlg:
            period_time_bounds = period_cfg.time_bounds
            start_time_matches = dataset_time_bounds[0] >= period_time_bounds[0]
            end_time_matches = dataset_time_bounds[0] <= period_time_bounds[1]
            if start_time_matches and end_time_matches:
                mmsis = period_cfg.mmsis
        return mmsis

    def get_floenavi_cfg(self, dataset_time_bounds):
        """
        Searches for the mmsi number of the reference and heading station
        :param dataset_time_bounds: (datetime list) The start and end time of the dataset
        :return:
        """

        # Check time coverage for each FloeNavi configuration period
        # (period section in the floenavi yaml config file)
        floenavi_cfg = None
        for period_cfg in self.period_ctlg:
            period_time_bounds = period_cfg.time_bounds
            start_time_matches = dataset_time_bounds[0] >= period_time_bounds[0]
            end_time_matches = dataset_time_bounds[0] <= period_time_bounds[1]
            if start_time_matches and end_time_matches:
                floenavi_cfg = period_cfg

        return floenavi_cfg

    def get_dship_id(self, mmsi):
        """
        Return the dship id
        :param mmsi:
        :return:
        """
        return self.mmsi_ctlg.get_basestation_dship_id(mmsi)

    def get_basestations_ctlg_entries(self, mmsis=None, period=None):
        """
        Return a list of all catalog entries fitting optional search criteria
        :param mmsis: (int) optional list of mmsis. If None, all mmsi's will be used
        :param period: (list of datetime.date) optional period, If None, full period will be used
        :return: list of FloeNaviBaseStation instances
        """

        # Use full mmsi range, if no search criteria for mmsi has been specified
        if mmsis is None:
            mmsis = self.mmsi_ctlg.base_mmsis

        # Use full mmsi range, if no search criteria for period has been specified
        if period is None:
            period = self.refstations_ctlg.full_date_range

        # Get the list of days from the time bounds of the data set
        # The list of days is already in format contained in the filename
        tdelta = period[1]-period[0]
        n_days = tdelta.days
        date = period[0]
        date_strs = []
        for day_num in np.arange(n_days+1):
            next_date = date + timedelta(days=int(day_num))
            date_strs.append(next_date.strftime("%Y%m%d"))
        logger.debug("Period requires base station data from %g day(s)" % len(date_strs))

        # Return file(s) as a list as it could be more than one
        stations = []
        for date_str in date_strs:
            station_list = self.refstations_ctlg.get_station_entries(date_str, mmsi_filter=mmsis)
            stations.extend(station_list)

        # All done, return
        return stations

    def get_basestation_files(self, mmsis=None, period=None):
        """
        Get the list of file path
        :param mmsis: (int) optional list of mmsis. If None, all mmsi's will be used
        :param period: (list of datetime.date) optional period, If None, full period will be used
        :return: list of all filepaths
        """

        # Use full mmsi range, if no search criteria for mmsi has been specified
        if mmsis is None:
            mmsis = self.mmsi_ctlg.base_mmsis

        # Use full mmsi range, if no search criteria for period has been specified
        if period is None:
            period = self.refstations_ctlg.full_date_range

        # Get the list of days from the time bounds of the data set
        # The list of days is already in format contained in the filename
        tdelta = period[1]-period[0]
        n_days = tdelta.days
        date = period[0]
        date_strs = [date.strftime("%Y%m%d")]
        for day_num in np.arange(n_days):
            next_date = date + timedelta(days=int(day_num))
            date_strs.append(next_date.strftime("%Y%m%d"))
        logger.debug("Period requires base station data from %g day(s)" % len(date_strs))

        # Return file(s) as a list as it could be more than one
        station_filepaths = []
        for date_str in date_strs:
            file_paths = self.refstations_ctlg.get_station_filepath(date_str, mmsi_filter=mmsis)
            station_filepaths.extend(file_paths)

        # All done, return
        return station_filepaths

    def _parse_yaml(self):
        """
        Read the FloeNavi config file and map contents to a dictionary
        :return:
        """
        logger.debug("Read FloeNavi config file: %s" % os.path.abspath(self.yaml_config_file))
        with open(self.yaml_config_file) as fh:
            self._cfg = yaml.load(fh, Loader=yaml.FullLoader)

        logger.debug("Read FloeNavi path definition file: %s" % os.path.abspath(self.local_path_file))
        with open(self.local_path_file) as fh:
            self._pathdef = yaml.load(fh, Loader=yaml.FullLoader)

    def _create_valid_mmsi_catalog(self):
        """
        Get the valid mmsi's for base stations and mobile stations
        :return:
        """
        self._mmsi_ctlg = FloeNaviMMSICtlg(self._cfg["mmsis"])

    def _create_reference_data_catalog(self):
        """
        Create a catalog of data files (for each mmsi and day)
        :return:
        """

        # Find reference station files
        filename_lookup = self._pathdef["filename_template"].format(mmsi="*", yyyymmdd="*")
        filepath_lookup = os.path.join(self._pathdef["lookup_dir"], filename_lookup)
        refstat_files = sorted(glob.glob(filepath_lookup))
        msg = "Found %g %s files in %s"
        msg = msg % (len(refstat_files), self._pathdef["data_source_id"], self._pathdef["lookup_dir"])
        logger.info(msg)

        # Add file to catalog
        self._ref_ctlg = RefDataFileCatalog(refstat_files, self._pathdef["filename_template"])

    def _create_period_catalog(self):
        """
        Create a catalog of periods with defined FloeNavi configurations
        (mmsi of deployed basestations and their location on the FloeNavi grid
        :return:
        """
        self._prd_ctlg = []
        for period_cfg in self._cfg["periods"]:
            self._prd_ctlg.append(FloeNaviBaseStationSetup(period_cfg))

    @property
    def mmsi_ctlg(self):
        return self._mmsi_ctlg

    @property
    def period_ctlg(self):
        return self._prd_ctlg

    @property
    def refstations_ctlg(self):
        return self._ref_ctlg

    @property
    def basestation_ids(self):
        return self.mmsi_ctlg.base_mmsis

    @property
    def mobiletation_ids(self):
        return self.mmsi_ctlg.mobile_mmsis

    @property
    def pathdef(self):
        return dict(self._pathdef)

    @property
    def date_coverage_start(self):
        return self.refstations_ctlg.full_date_range[0].date()

    @property
    def date_coverage_end(self):
        return self.refstations_ctlg.full_date_range[1].date()


class FloeNaviMMSICtlg(object):
    """
    Container for the list of valid MMSI for basestations and mobile stations
    """

    def __init__(self, cfg_dict):
        """

        :param cfg_dict:
        """

        # save args
        self._cfg_dict = cfg_dict

        # Get dicts for base stations and mobile stations: {mmsi: labels}
        self._basestations = {}
        for item in self._cfg_dict["basestations"]:
            self._basestations[item["mmsi"]] = BaseStationDef(item)

        self._mobilestations = {}
        for item in self._cfg_dict["mobilestations"]:
            self._mobilestations[item["mmsi"]] = BaseStationDef(item)

    def get_type(self, mmsi):
        """
        Return the type (base or mobile) of a given mmsi, None if invalid mmsi
        :param mmsi:
        :return:
        """
        if mmsi in self.base_mmsis:
            station_type = "base"
        elif mmsi in self.mobile_mmsis:
            station_type = "mobile"
        else:
            station_type = None
        return station_type

    def get_station_name(self, mmsi):
        """
        Return the station name for a given mmsi, None if invalid mmsi
        :param mmsi:
        :return:
        """
        if mmsi in self.base_mmsis:
            station_name = self._basestations[mmsi].label
        elif mmsi in self.mobile_mmsis:
            station_name = self._mobilestations[mmsi].label
        else:
            station_name = None
        return station_name

    def get_basestation_dship_id(self, mmsi):
        """
        Return the DSHIP id (
        :param mmsi:
        :return:
        """
        # DShips ids only available for base stations
        if mmsi not in self.base_mmsis:
            logger.error("MMSI %d not a registered basestation -> no DSHIP id exists")
            return None
        return self._basestations[mmsi].dship_id

    @property
    def base_mmsis(self):
        return sorted(self._basestations.keys())

    @property
    def mobile_mmsis(self):
        return sorted(self._mobilestations.keys())


class BaseStationDef(object):
    """
    Small data class for the mmsi station definition in the config file
    Attributes:
        - mmsi
        - label
        - dship_id (optional)
    """

    def __init__(self, cfg_dict):
        """
        Create a basestation definition data class
        :param cfg_dict:
        """
        self.mmsi = cfg_dict["mmsi"]
        self.label = cfg_dict["label"]
        self.dship_id = cfg_dict.get("dship_id", None)


class FloeNaviBaseStationSetup(object):
    """
    The configuration of the basestations (mmsi and their positions)
    in the ice coordinate system (FloeNavi Grid)
    """

    def __init__(self, period_cfg_dict):
        """

        :param period_cfg_dict:
        """
        self.period_cfg_dict = period_cfg_dict

    def get_grid_location(self, mmsi):
        """
        Get the reference position of a basestation within the
        FloeNavi grid
        :param mmsi:
        :return:
        """
        for cfg in self.period_cfg_dict["basestations"]:
            if int(cfg["mmsi"]) == mmsi:
                return cfg["pos"]
        return None, None

    @property
    def time_bounds(self):
        tcs = self.period_cfg_dict["valid_since"]
        tce = self.period_cfg_dict["valid_until"]
        if tce is None:
            tce = datetime.now()
        return [tcs, tce]

    @property
    def mmsis(self):
        return [entry["mmsi"] for entry in self.period_cfg_dict["basestations"]]

    @property
    def icecs_trns_kwargs(self):
        return self.period_cfg_dict["icecs_transformation"]

    @property
    def cfg_num(self):
        return self.period_cfg_dict["cfg_num"]


class RefDataFileCatalog(object):
    """
    Collection of metadata for all files in the basestation data folder
    """

    def __init__(self, filepaths, parser):

        self.filepaths = filepaths
        self._catalog = []
        for filepath in self.filepaths:
            try:
                catalog_item = RefDataCatalogItem.from_filepath(filepath, parser)
            except ValueError:
                logger.error("Unable to parse file: %s -> skipping file" % filepath)
                continue
            self._catalog.append(catalog_item)

        # Check results
        # logger.info("MMSI's in catalog:")
        # for mmsi, start_time, stop_time in self.mssi_coverage:
        #     msg = "mmsi: %d [%s - %s]"
        #     msg = msg % (mmsi, start_time.isoformat(), stop_time.isoformat())
        #     logger.info(msg)

    def get_items(self, mmsi):
        return [item for item in self._catalog if item.mmsi == mmsi]

    def get_station_entries(self, yyyymmdd, mmsi_filter=None):
        """
        Get filepaths for files for a given date and mmsi (potentially more than one)
        :param yyyymmdd:
        :param mmsi_filter:
        :return:
        """
        if mmsi_filter is None:
            mmsi_filter = self.mmsis
        if isinstance(mmsi_filter, int):
            mmsi_filter = [mmsi_filter]
        filepaths = [item for item in self._catalog if item.yyyymmdd == yyyymmdd and item.mmsi in mmsi_filter]
        return filepaths

    def get_station_filepath(self, yyyymmdd, mmsi_filter=None):
        """
        Get filepaths for files for a given date and mmsi (potentially more than one)
        :param yyyymmdd:
        :param mmsi_filter:
        :return:
        """
        if mmsi_filter is None:
            mmsi_filter = self.mmsis
        if isinstance(mmsi_filter, int):
            mmsi_filter = [mmsi_filter]
        filepaths = [item.filepath for item in self._catalog if item.yyyymmdd == yyyymmdd and item.mmsi in mmsi_filter]
        return filepaths

    @property
    def mmsis(self):
        return set(sorted([item.mmsi for item in self._catalog]))

    @property
    def mssi_coverage(self):
        # TODO: Refactor to a list that also contains the gaps between the first and last day
        mssi_coverage = []
        for mmsi in self.mmsis:
            items = self.get_items(mmsi)
            start_times = [item.time_bounds[0] for item in items]
            stop_times = [item.time_bounds[1] for item in items]
            mssi_coverage.append([mmsi, np.amin(start_times), np.amax(stop_times)])
        return mssi_coverage

    @property
    def full_date_range(self):
        start_times = []
        stop_times = []
        for mmsi in self.mmsis:
            items = self.get_items(mmsi)
            start_times.extend([item.time_bounds[0] for item in items])
            stop_times.extend([item.time_bounds[1] for item in items])
        return [np.amin(start_times), np.amax(stop_times)]


class RefDataCatalogItem(object):
    """
    Container for metadata for one AIS stations (basestation of mobilestation) file
    """

    def __init__(self, filepath, mmsi=None, yyyymmdd=None, time_coverage="daily"):
        self.filepath = filepath
        self.mmsi = int(mmsi)
        self.time_coverage = time_coverage
        self.yyyymmdd = yyyymmdd
        self.ref_time = datetime.strptime(yyyymmdd, "%Y%m%d")
        self.time_bounds = [
            datetime(self.ref_time.year, self.ref_time.month, self.ref_time.day, 0, 0, 0),
            datetime(self.ref_time.year, self.ref_time.month, self.ref_time.day, 23, 59, 59)]

    @classmethod
    def from_filepath(cls, filepath, parser):

        # Get parameter from the filename
        filename = os.path.split(filepath)[-1]
        result = parse.parse(parser, filename)
        attrs = dict()
        if result:
            for key in result.named.keys():
                attrs[key] = result.named[key]
        return cls(filepath, **attrs)

    @property
    def filename(self):
        return os.path.split(self.filepath)[-1]

    @property
    def date(self):
        return self.ref_time.date()


class FloeNaviGrid(object):
    """
    A container for all basestation data in a defined period
    and functionality to compute two virtual basestation
    that define heading and origin for the ice coordindate system
    based on two or more FloeNavi reference stations
    """

    def __init__(self, cfg, setup):
        """

        :param cfg:
        :param setup:
        """

        # Store the variables
        self.cfg = cfg
        self.setup = setup

        # Attributes
        self.time_bounds = None
        self.basestations = None
        self.basestation_dict = {}

    def set_time_bounds(self, dataset_time_bounds, day_pad=True):
        """
        Specify the time bounds of the data set, which will trigger parsing the appropriate
        basestation data.
        :param dataset_time_bounds:
        :param day_pad:
        :return:
        """

        # if day_pad is set to True: Extend the range of the floenavi grid
        # by two padding days (if the setup allows)
        # NOTE: This can improve the solution right at the start and end of an UTC day
        time_bounds = list(dataset_time_bounds)
        if day_pad:
            early_start_date = time_bounds[0] - timedelta(days=1)
            if early_start_date >= self.setup.time_bounds[0]:
                time_bounds[0] = early_start_date
            later_end_date = time_bounds[1] + timedelta(days=1)
            if later_end_date <= self.setup.time_bounds[1]:
                time_bounds[1] = later_end_date

        # Store the current time_bounds
        self.time_bounds = time_bounds

        # TODO: Narrow list of mmsi to one specified in config file (not all available in repo)

        # Get the list of base station files (yet unsorted)
        basestations_ctlg_entries = self.cfg.get_basestations_ctlg_entries(
            mmsis=self.setup.mmsis, period=time_bounds)

        # Load the base station data and append data for specific mmsi's
        # This step is there to group data by mmsi in case there more than
        # one file per mmsi
        self.basestation_dict = {}
        for basestation in basestations_ctlg_entries:
            mmsi = basestation.mmsi
            logger.info("Read Reference station file(s): %s" % basestation.filepath)
            ais_data = DshipAISExtract(basestation.filepath)
            pos = GeoPositionData(ais_data.time, ais_data.longitude, ais_data.latitude)
            if mmsi not in self.basestation_dict:
                self.basestation_dict[mmsi] = pos
            else:
                self.basestation_dict[mmsi].append(pos)

    def get_virtual_reference_station_positions(self, time_bounds, time_resolution_sec=10, exit_on_error=True,
                                                **refstat_kwargs):
        """
        Compute the position objects of two virtual reference stations, based on the number of available
        basestations. This is primary input for the icedrift correction in the icedrift module.
        :param time_resolution_sec: (float) The temporal resolution of the virtual base station positions
        :param time_bounds: (datetime array) The time bounds of the required data period
            (might differ from the available time bounds for the basestation setuo)
        :param exit_on_error: (bool) Flag if to sys.exit() on error or not
        :param refstat_kwargs: (dict) keywords for the reference station computation
        :return:
        """

        # Compute common timestamp
        logger.info("Compute timestamp of virtual base stations [resolution: %.2f sec]" % time_resolution_sec)
        duration_seconds = (time_bounds[1]-time_bounds[0]).seconds
        seconds_increments = np.arange(0, duration_seconds+time_resolution_sec, time_resolution_sec)
        time = np.array([time_bounds[0]+timedelta(seconds=int(sec)) for sec in seconds_increments])

        # Create a collection of all base stations
        # NOTE: Making sure all base stations have the same temporal resolutions
        logger.info("Create base station collection with harmonized timestamp")
        self.basestations = FloeNaviBaseStationCollection()
        for mmsi in self.basestation_dict:
            basestation_pos = self.basestation_dict[mmsi]
            basestation_pos.resample_to_time(time, max_dist2data_secs=300)
            ref_pos = self.setup.get_grid_location(mmsi)
            self.basestations.add_basestation(FloeNaviBaseStation(mmsi, ref_pos, basestation_pos))

        # Pass information to grid solver and compute origin and heading from
        # all base stations
        logger.info("Compute FloeNavi grid solution from basestation collection")
        fngrid = FloeNaviGridSolver(self.basestations, setup=self.setup)
        try:
            fngrid.solve()
            # Get the virtual origin and heading position data objects
            logger.info("Compute origin and heading station positions")
            orig, hdg = fngrid.get_reference_stations(**refstat_kwargs)
        except SystemExit as sysexit:
            if exit_on_error:
                raise sysexit
            logger.warning("No origin heading stations availabel for selected period")
            orig, hdg = None, None
            logger.warning("")

        return orig, hdg, fngrid


class FloeNaviGridSolver(object):
    """
    Class for managing and obtaining the origin of the FloeNavi grid in geographical
    coordinates and heading of the x-axis from an arbitrary set of deployed AIS
    basestations throughout the floe.

     This algorithm
    is implemented in the FloeNaviGridSolution class.

    This class will repeat the computation over all unique permutations of basestations
    and provides a merged solution with optional smoothing and filtering.

    """

    def __init__(self, basestations, setup=None, minimum_distance_m=100., smoothing_window=None):
        """
        Init the FloeNavi Grid solver
        :param basestations:
        :param setup:
        :param minimum_distance_m:
        :param smoothing_window:
        """

        # Input arguments
        self.basestations = basestations
        self.setup = setup
        self.minimum_distance_m = minimum_distance_m
        self.smoothing_window = smoothing_window

        # Attributes
        self.solutions = []
        self.origin_longitude = None
        self.origin_latitude = None
        self.origin_uncertainty_m = None
        self.axes_heading = None
        self.heading_uncertainty_deg = None

    def solve(self):
        """
        Compute the multi base station solution for the origin and axes heading of the
        FloeNavi grid from all possible permutations of the base station combinations
        """

        # Get permutations
        mmsi_permutations = self.mmsi_permutations
        n_permutations = len(mmsi_permutations)
        logger.info("Base stations combinations available for FloeNavi grid solution: %d" % n_permutations)

        # Input validation
        if n_permutations == 0:
            logger.error("To few basestations available -> Terminating")
            sys.exit(1)

        # Loop over all permutations
        for mmsi1, mmsi2 in self.mmsi_permutations:

            # Get base station data objects
            logger.info("Using combination: [%d -> %d]" % (mmsi1, mmsi2))
            bs1 = self.basestations.get_by_mmsi(mmsi1)
            bs2 = self.basestations.get_by_mmsi(mmsi2)

            # Check distance between the two base stations
            bs_distance_m = self.basestations_grid_distance(bs1, bs2)
            if bs_distance_m >= self.minimum_distance_m:
                msg = "Distance between stations: %.0fm [min: %.0fm -> baseline ok]"
                msg = msg % (bs_distance_m, self.minimum_distance_m)
                logger.info(msg)
            if bs_distance_m < self.minimum_distance_m:
                msg = "Distance between stations: %.0fm [min: %.0fm -> baseline to short]"
                msg = msg % (bs_distance_m, self.minimum_distance_m)
                logger.warning(msg)
                continue

            # Compute the solutions and add to the list
            self.solutions.append(FloeNaviGridSolution(bs1, bs2))

        # --- Check result ---

        # No solution -> error
        if self.n_solutions == 0:
            logger.error("No FloeNavi grid solution available -> terminating")
            sys.exit(1)

        # One solution -> return as is
        if self.n_solutions == 1:
            logger.info("Single solution, skipping merging")
            solution = self.solutions[0]
            self.origin_longitude = solution.origin_longitude
            self.origin_latitude = solution.origin_latitude
            self.origin_uncertainty_m = solution.origin_uncertainty_m
            self.axes_heading = solution.axes_heading
            return

        # Multiple solutions -> merge
        logger.info("Merge %d FloeNavi grid solutions" % self.n_solutions)

        # NOTE: Currently, this is a simple point-wise mean of the parameters
        # TODO: Add quality check (e.g. the min-max distance)
        parameter_names = ["origin_longitude", "origin_latitude", "axes_heading", "origin_uncertainty_m"]
        parameter_units = dict(origin_longitude="deg", origin_latitude="deg", origin_uncertainty_m="m",
                               axes_heading="deg")
        n_records = self.solutions[0].n_records
        for parameter_name in parameter_names:

            # Merge solution for parameters in to a multi-dimensional array
            # -> easier to calculate mean and statistics with numpy
            var_array = np.ndarray(shape=(n_records, self.n_solutions))
            for solution_num in np.arange(self.n_solutions):
                var = getattr(self.solutions[solution_num], parameter_name)
                if var is None:
                    var = np.full([n_records], np.nan)
                var_array[:, solution_num] = var

            # Report statistics based on spread of solutions
            unit = parameter_units[parameter_name]
            min_vals = np.nanmin(var_array, axis=1)
            max_vals = np.nanmax(var_array, axis=1)
            spread = max_vals - min_vals
            msg = "Spread of solutions for %s: max: %f%s, sdev:%f%s"
            msg = msg % (parameter_name, np.nanmax(spread), unit, np.nanstd(spread), unit)
            logger.debug(msg)

            # Assign the mean solution (this includes the origin uncertainty
            mean_solution = np.nanmean(var_array, axis=1)
            setattr(self, parameter_name, mean_solution)

            # Compute heading uncertainty (point-wise standard deviation of different solutions)
            if parameter_name == "axes_heading":
                self.heading_uncertainty_deg = np.nanstd(var_array, axis=1)

            # Compute origin uncertainty (point-wise mean of different solutions)
            if parameter_name == "origin_uncertainty_m":
                self.origin_uncertainty_m = np.nanmean(var_array, axis=1)

    @staticmethod
    def basestations_grid_distance(bs1, bs2):
        return np.sqrt((bs2.xc_grid-bs1.xc_grid)**2.0 + (bs2.yc_grid-bs1.yc_grid)**2.0)

    def get_reference_stations(self, heading_station_distance_m=500):
        """
        Returns two position object (one for origin and one for specifying the heading of the primary
        axis) that can be used as input for the icedrift package.
        :return: origin, hdg: Position objects for origin and heading of the FloeNavi x-axis
        """

        # Origin stations
        origin = GeoPositionData(self.basestations.time, self.origin_longitude, self.origin_latitude)

        # Compute heading station positions
        hdg = origin.clone()
        az = self.axes_heading
        r = heading_station_distance_m
        g = pyproj.Geod(ellps='WGS84')
        for i in np.arange(hdg.n_records):
            hdg.longitude[i], hdg.latitude[i], _ = g.fwd(hdg.longitude[i], hdg.latitude[i], az[i], r)

        return origin, hdg

    @property
    def mmsis(self):
        return self.basestations.mmsis

    @property
    def mmsi_permutations(self):
        """
        Return all permutations of base stations that can be used to solve for grid origin
        and location
        :return:
        """
        mmsi_permutations = []
        values = self.mmsis
        all_values = list(self.mmsis)
        for entry in all_values:
            values.pop(values.index(entry))
            for other_entry in values:
                mmsi_permutations.append((entry, other_entry))
        return mmsi_permutations

    @property
    def basestation_distance_error(self):
        """
        Computes the average basestation distance error
        -> Mean offset between expected and measured basestation distance for all solutions
        :return:
        """
        mean_distance_error = np.ndarray(shape=(self.n_records, self.n_solutions))
        for snum in np.arange(self.n_solutions):
            mean_distance_error[:, snum] = np.abs(getattr(self.solutions[snum], "bs_measured_distance_offset"))
        return np.nanmean(mean_distance_error, axis=1)

    @property
    def n_solutions(self):
        return len(self.solutions)

    @property
    def n_records(self):
        return self.solutions[0].n_records


class FloeNaviGridSolution(object):
    """
    Class containing the algorithm for computing origin and heading from FloeNavi
    basestations at arbitrary (but known) positions in the FloeNavi Grid.

    The main idea of the algorithm is to compute the azimuth of the FloeNavi x-axis by
    comparing the measured azimuth between two base stations against the defined azimuth
    in the FloeNavi grid coordinate system. The origin of the FloeNavi grid in
    geographical coordinates than can be computed for the known positions of the
    base station in the FloeNavi grid and the heading of the x-axis.
    """

    def __init__(self, bs1, bs2, distance_deviation_warning_m=20.0):
        """
        Compute origin and heading of the FloeNavi grid based on the geographical
        position of two base station with known positions in the grid
        :param bs1: Basestation 1 (floenavi.FloeNaviBaseStation)
        :param bs2: Basestation 2 (floenavi.FloeNaviBaseStation)
        :param distance_deviation_warning_m: (float) Warning threshold for discrepancies between observed
            and specified base station distances
        """

        # Store arguments
        self.bs1 = bs1
        self.bs2 = bs2
        self.distance_deviation_warning_m = distance_deviation_warning_m

        # Init attributes
        self.azimuth_stations_geo = None
        self.azimuth_stations_grid = None
        self.origin_longitude = None
        self.origin_latitude = None
        self.origin_uncertainty_m = None
        self.axes_heading = None
        self.bs_measured_distance = None
        self.bs_measured_distance_offset = None

        # Compute origin and heading
        self._compute()

    def _compute(self):
        """
        Get origin in geographical coordinates and heading of x-axis of the
        FloeNavi grid
        :return:
        """
        # Init the stereographic projection and compute geographical x,y
        # positions of base stations
        logger.info("Compute projection based on position of %d" % self.bs1.mmsi)
        proj_dict = self.proj_dict
        logger.debug(str(proj_dict))
        for bs in [self.bs1, self.bs2]:
            bs.set_projection(proj_dict)

        # Compute and compare base station distances with expected values
        self._compute_basestation_distances()

        # Compute the heading of the axis by evaluation true heading between base
        # stations with expected values
        self._compute_axes_heading()

        # Compute the origin in projection coordinates
        # NOTE: This can be done for each base station
        self._compute_origin_coordinates()

    def _compute_basestation_distances(self):
        """
        Compute the measured distance between the two basestations and compare the result
        with the expected value
        :return:
        """

        # Compute average distance between base stations and compare to specified value
        bs_true_distance = np.sqrt((self.bs2.geo.xc-self.bs1.geo.xc)**2. + (self.bs2.geo.yc-self.bs1.geo.yc)**2.)
        bs_grid_distance = np.sqrt((self.bs2.xc_grid-self.bs1.xc_grid)**2. + (self.bs2.yc_grid-self.bs1.yc_grid)**2.)

        # Check for NaN's
        invalid = np.where(bs_true_distance > 1.0e10)[0]
        bs_true_distance[invalid] = np.nan

        average_true_distance = np.nanmean(bs_true_distance)
        msg = "Observed base station distance: %.1f m [specified: %.1f m]"
        logger.debug(msg % (average_true_distance, bs_grid_distance))
        average_distance_deviation = bs_grid_distance - average_true_distance
        if average_distance_deviation > self.distance_deviation_warning_m:
            msg = "High discrepancy between observed and specified base station distance: %.1f m"
            logger.warning(msg % average_distance_deviation)

        # Save as property
        self.bs_measured_distance = bs_true_distance
        self.bs_measured_distance_offset = bs_true_distance - bs_grid_distance

    def _compute_axes_heading(self):
        """
        Compute the heading of the grid axes as offset between measured and expected
        true heading between the base stations
        :return:
        """

        # Test
        old_azimuth_stations_geo = self.angle_between_points(self.bs1.geo.xc, self.bs1.geo.yc,
                                                             self.bs2.geo.xc, self.bs2.geo.yc)

        # Compute the geographical azimuth between the two base stations
        g = pyproj.Geod(ellps='WGS84')
        self.azimuth_stations_geo, _, _ = g.inv(
            self.bs1.geo.longitude, self.bs1.geo.latitude,
            self.bs2.geo.longitude, self.bs2.geo.latitude)

        self.azimuth_stations_geo = self.azimuth_stations_geo % 360.

        # Issue: the computation of the azimuth will provide finite values for infinite input
        # -> Double check the result
        invalid = np.logical_or(np.logical_not(np.isfinite(self.bs1.geo.longitude)),
                                np.logical_not(np.isfinite(self.bs2.geo.longitude)))
        self.azimuth_stations_geo[invalid] = np.nan

        msg = "Geographical azimuth of base stations: %.1f deg [min: %.1f, max: %.1f]"
        msg = msg % (np.nanmean(self.azimuth_stations_geo),
                     np.nanmin(self.azimuth_stations_geo),
                     np.nanmax(self.azimuth_stations_geo))
        logger.debug(msg)

        # Compute the reference azimuth of the two base stations in FloeNavi grid coordinates
        self.azimuth_stations_grid = self.angle_between_points(self.bs1.xc_grid, self.bs1.yc_grid,
                                                               self.bs2.xc_grid, self.bs2.yc_grid)
        logger.debug("Target azimuth of base stations: %.1f deg" % self.azimuth_stations_grid)

        # Compute the heading of the x-axis
        # NOTE: Reference axes for FloeNavi grid is the x-axis
        logger.info("Compute heading of FloeNavi x-axis")
        self.axes_heading = self.azimuth_stations_geo + self.azimuth_stations_grid
        self.axes_heading = self.axes_heading % 360.0

        # The rotation is then bringing the heading of the axes to the 90deg position
        # !!!: Needs evaluation
        self.rotation = self.axes_heading.copy()

    def _compute_origin_coordinates(self):
        """
        Computing the origin of the grid in geographic coordinates by evaluating the
        two basestations positions in a grid with a known true heading (previous step)
        :return:
        """

        origins_xc = np.full((2, self.bs1.n_records), np.nan)
        origins_yc = origins_xc.copy()
        for i, bs in enumerate([self.bs1, self.bs2]):
            xc, yc = self.get_origin(bs.geo.longitude, bs.geo.latitude, bs.range_m, bs.ref_angle, self.rotation)
            invalid = np.where(np.logical_or(xc > 1.0e10, yc > 1.0e10))[0]
            xc[invalid], yc[invalid] = np.nan, np.nan
            origins_xc[i, :] = xc
            origins_yc[i, :] = yc

        # import matplotlib.pyplot as plt
        # plt.figure()
        # for i, bs in enumerate([self.bs1, self.bs2]):
        #     # plt.scatter(bs.geo.xc, bs.geo.yc, s=40, label="geo {}".format(bs.mmsi))
        #     plt.scatter(origins_xc[i, :], origins_yc[i, :], s=5, label="origin {}".format(bs.mmsi))
        # plt.legend()
        # plt.show()

        # Compute the average deviation (uncertainty) of the origin
        origin_xc_delta = origins_xc[1]-origins_xc[0]
        origin_yc_delta = origins_yc[1]-origins_yc[0]
        self.origin_uncertainty_m = np.sqrt(origin_xc_delta**2.0 + origin_yc_delta**2.0)
        invalid = np.where(self.origin_uncertainty_m > 1.0e10)[0]

        self.origin_uncertainty_m[invalid] = np.nan
        logger.debug("Average origin uncertainty: %.1fm" % np.nanmean(self.origin_uncertainty_m))

        # Compute the origin in geographical coordinates
        logger.info("Compute FloeNavi origin in lon/lat")
        p = pyproj.Proj(**self.proj_dict)
        origin_xc = np.nanmean(origins_xc, axis=0)
        origin_yc = np.nanmean(origins_yc, axis=0)
        self.origin_longitude, self.origin_latitude = p(origin_xc, origin_yc, inverse=True)

    @staticmethod
    def angle_between_points(xc1, yc1, xc2, yc2, deg_output=True):
        """
        Compute the angle in cartesian coordinates between two points
        :param xc1:
        :param yc1:
        :param xc2:
        :param yc2:
        :param deg_output:
        :return:
        """

        # Angle with respect to x-axis
        angle = np.arctan2(yc2 - yc1, xc2 - xc1)
        angle = np.mod(angle, 2.0*np.pi)

        # Convert output (optional) and return
        if deg_output:
            angle = np.rad2deg(angle)
        return angle

    def get_origin(self, lon, lat, range_m, ref_heading, rotation):
        """
        Compute the origin from basestation data and the known rotation of the
        ice reference system
        :param xc:
        :param yc:
        :param range_m:
        :param ref_heading:
        :param rotation:
        :return:
        """

        # Step 1: Project the origin positions in geographic space
        direction_to_origin_deg = rotation - ref_heading + 180.
        g = pyproj.Geod(ellps='WGS84')
        origin_lon, origin_lat, _ = g.fwd(lon, lat, direction_to_origin_deg, np.full(lon.shape, range_m))

        # Step 2: Compute x, y in local projection coordinates
        p = pyproj.Proj(**self.proj_dict)
        xc_orig, yc_orig = p(origin_lon, origin_lat)

        return xc_orig, yc_orig

    @property
    def proj_dict(self):
        lon_0 = np.nanmedian(self.bs1.geo.longitude)
        lat_0 = np.nanmedian(self.bs1.geo.latitude)
        proj_dict = dict(proj='stere', lon_0=lon_0, lat_0=lat_0, lat_ts=lat_0, ellps='WGS84')
        return proj_dict

    @property
    def n_records(self):
        return len(self.origin_longitude)


class FloeNaviBaseStationCollection(object):
    """
    A container for all basestation data and their metadata in
    a specific FloeNavi configuration
    """

    def __init__(self, basestations=None):
        """
        Init the list of basestations
        """
        if basestations is None:
            self.list = []
        else:
            self.list = basestations

    def add_basestation(self, basestation):
        """
        Add a basestation to the collection
        :param basestation: floenavi.FloeNaviBaseStation instance
        :return:
        """
        # TODO: input validation
        self.list.append(basestation)

    def get_by_mmsi(self, mmsi, error_on_invalid=False):
        """
        Return basestation data based on the mmsi
        :param mmsi:
        :param error_on_invalid:
        :return:
        """
        basestations = [basestation for basestation in self.list if basestation.mmsi == mmsi]
        if len(basestations) == 0:
            basestations = None
        elif len(basestations) == 1:
            basestations = basestations[0]
        else:
            if error_on_invalid:
                raise ValueError("No basestation with mmsi: %d" % mmsi)
        return basestations

    @property
    def mmsis(self):
        return list(sorted([basestation.mmsi for basestation in self.list]))

    @property
    def time(self):
        if self.n_basestations == 0:
            return None
        basestation = self.list[0]
        return basestation.geo.time

    @property
    def n_basestations(self):
        return len(self.list)


class FloeNaviBaseStation(object):
    """
    Container for FloeNavi basestation data including observed
    geographic locations and known location on the FloeNavi grid
    """

    def __init__(self, mmsi, ref_pos, geo):
        """
        Init the basestation data
        :param mmsi:
        :param ref_pos:
        :param geo:
        """

        # Store Attributes
        self.mmsi = mmsi
        self.ref_pos = ref_pos
        self.geo = geo

    def set_projection(self, proj_dict):
        """
        Compute geographical cartesian coordinates with a given projection
        (Usually polar-stereographic)
        :param proj_dict:
        :return:
        """
        p = pyproj.Proj(**proj_dict)
        self.geo.xc, self.geo.yc = p(self.geo.longitude, self.geo.latitude)

    @property
    def xc_grid(self):
        return self.ref_pos[0]

    @property
    def yc_grid(self):
        return self.ref_pos[1]

    @property
    def range_m(self):
        return np.sqrt(self.ref_pos[0]**2. + self.ref_pos[1]**2.)

    @property
    def ref_angle(self):
        return np.rad2deg(np.arctan2(self.ref_pos[1], self.ref_pos[0]))

    @property
    def n_records(self):
        return self.geo.n_records
