# -*- coding: utf-8 -*-

"""
This module contains dataset classes intended for the use with the FloeNavi().get_xy_coordinates()
method.

Convention for dataset classes:

    Required attributes
    - time: array of datetime.datetime in UTC (or at least in the same timezone as the base stations)
    - longitude: float array
    - latitude: float array
    - xc: float array or None if not yet computed
    - yc: float array or None if not yet computed
    - n_records: Number of records
    - time_bounds: [time_coverage_start, time_coverage_end] as datetime.datetime objects

    Required methods:
    - update_floenavi_coords(self, xc, yc): Setting of (x,y) by FloeNavi().get_xy_coordinates()

"""


import sys
from loguru import logger
import numpy as np
from matplotlib.dates import date2num
from dateutil.parser import parse as du_parse


class Waypoints(object):
    """
    FloeNavi dataset for GPS waypoints data with label, time, longitude, latitude. To meant be used as

        wps = Waypoints()
        for label, time, lon, lat in <some-iterable>:
            add(label, time, lon, lat)
        floenavi = FloeNavi()
        floenavi.get_xy_coordinates(wps)

    After succesful transformation, the waypoint instance will have xc, yc attributes.

    The class can also be initialized from a csv file with the class method:

        wps = Waypoints.from_csv(filename)

    given that the input file has comma-separated columns for label, time, longitude, latitude:

    Example:

        081-Polarstern, 2019-10-06T01:43:20, 133.780012987554073, 85.109757967293262
        082-OceanCity-PowerHub, 2019-10-06T02:04:43, 133.748755026608706, 85.110806962475181
        083-Ridge1, 2019-10-06T02:13:21, 133.747596982866526, 85.110585009679198

    """

    def __init__(self):
        """
        Create a Waypoint instance. Waypoint data has to be added with the method `add(label, time, lon, lat)`
        """
        self._wps = []

    @classmethod
    def from_csv(cls, filepath):
        """
        Read a standard waypoint csv file with 4 columns:

            waypoint label (str), datetime (str), latitude (float), longitude (float)

        :param filepath:
        :return:
        """

        # --- Read the csv content ---
        names = ('label', 'time', 'latitude', 'longitude')
        formats = ('U96', 'U96', 'f4', 'f4')
        content = np.loadtxt(filepath, skiprows=1, delimiter=',', dtype={'names': names, 'formats': formats})

        # Construct the instance and compute waypoints
        wps = cls()
        for label, time_str, lat, lon in content:
            time = du_parse(time_str)
            wps.add(label, time, lon, lat)

        # All done, return initialized instance
        return wps

    def add(self, *args):
        self._wps.append(WaypointData(*args))

    def update_floenavi_coords(self, xc, yc):
        """
        Add x, y coordinates to each waypoint
        :param xc:
        :param yc:
        :return:
        """
        for i, (xc_val, yc_val) in enumerate(zip(xc, yc)):
            self.wps[i].add_xy(xc_val, yc_val)

    @property
    def wps(self):
        return list(self._wps)

    @property
    def n_waypoints(self):
        return len(self.wps)

    @property
    def waypoints(self):
        return self.wps

    @property
    def time(self):
        return np.array([wp.time for wp in self.wps])

    @property
    def latitude(self):
        return np.array([wp.latitude for wp in self.wps])

    @property
    def longitude(self):
        return np.array([wp.longitude for wp in self.wps])

    @property
    def labels(self):
        return np.array([wp.label for wp in self.wps])

    @property
    def xc(self):
        return np.array([wp.xc for wp in self.wps])

    @property
    def yc(self):
        return np.array([wp.yc for wp in self.wps])

    @property
    def time_bounds(self):
        return [np.amin(self.time), np.amax(self.time)]


class GPSTrack(object):
    """
    FloeNavi dataset for track data with time, longitude, latitude. To meant be used as

        track = GPSTrack(times, longitude, latitude)
        floenavi = FloeNavi()
        floenavi.get_xy_coordinates(track)

    After succesful transformation, the track instance will have xc, yc attributes.

    The class can also be initialized from a csv file with the class method:

        wps = GPSTrack.from_csv(filename)

    given that the input file has comma-separated columns for time, longitude, latitude:

    Example:

        2019-10-24T05:22:18.433593, 129.20388794, 85.39130402
        2019-10-24T05:22:18.554687, 129.20387268, 85.39130402
        2019-10-24T05:22:18.675781, 129.20387268, 85.39130402

    """

    def __init__(self, time, longitude, latitude, xc=None, yc=None):
        """
        Create a GPS track data set for use with FloeNavi
        :param time: (datetime array) timestamp in UTC
        :param longitude: (float array) longitude values
        :param latitude: (float array) latitude values
        :param xc: (float array) Optional: FloeNavi x coordinates (will be set by FloeNavi)
        :param yc: (float array) Optional: FloeNavi y coordinates (will be set by FloeNavi)
        """

        # --- Store input ---
        self.time = time
        self.longitude = longitude
        self.latitude = latitude
        self.xc = xc
        self.yc = yc

    @classmethod
    def from_csv(cls, filepath, zero_latlon_to_nan=True, interpolate_gaps=True):
        """
        Read a standard waypoint csv file with 4 columns:

            waypoint label (str), datetime (str), latitude (float), longitude (float)

        :param filepath:
        :param zero_latlon_to_nan:
        :param interpolate_gaps:
        :return:
        """

        # --- Read the csv content ---
        names = ('time', 'latitude', 'longitude')
        formats = ('U96', 'f4', 'f4')
        content = np.loadtxt(filepath, skiprows=0, delimiter=',', dtype={'names': names, 'formats': formats})

        # Construct the instance and compute waypoints
        time = np.ndarray(shape=content.shape, dtype=object)
        for i, (time_str, _, _) in enumerate(content):
            time[i] = du_parse(time_str)

        # Sanity check and set invalid entries to NaN
        # NOTE: Currently, it is assumed that invalid entries have a value of 0.0
        # TODO: Generalize this to allow diverse nodata flags
        longitude, latitude = content["latitude"], content["longitude"]
        is_invalid = np.abs(latitude) < 1
        if zero_latlon_to_nan:
            invalid_indices = np.where(is_invalid)[0]
            longitude[invalid_indices] = np.nan
            latitude[invalid_indices] = np.nan

        # Simple gap interpolation
        if interpolate_gaps:
            datenum = date2num(time)
            valid_indices = np.where(np.logical_not(is_invalid))[0]
            longitude = np.interp(datenum, datenum[valid_indices], longitude[valid_indices])
            latitude = np.interp(datenum, datenum[valid_indices], latitude[valid_indices])

        # All done, return initialized instance
        track = cls(time, longitude, latitude)
        return track

    def update_floenavi_coords(self, xc, yc):
        """
        Add x, y coordinates to each waypoint
        :param xc:
        :param yc:
        :return:
        """

        # Simple Input sanity check (shape of arrays must match)
        for input_var in [xc, yc]:
            if input_var.shape != self.time.shape:
                msg = "Dimension mismatch: GPSTrack: %s != Coordinates: %s -> terminating"
                logger.error(msg % (str(self.time.shape), str(input_var.shape)))
                sys.exit(1)

        # Store (x,y) coordinates
        self.xc = xc
        self.yc = yc

    def export_csv(self, output_filepath):
        """
        Export the file as csv including xc & yc
        :param output_filepath:
        :return:
        """

        # Check if data set has assigned x, y coordinates
        # if not -> display warning and use warning
        if self.has_xy_coordinates:
            xc, yc = self.xc, self.yc
        else:
            xc, yc = np.full(self.time.shape, np.nan), np.full(self.time.shape, np.nan)
            logger.warning("Exporting GPSTrack without xc, yc: method `update_floenavi_coords` has not been used")

        # Create the track output
        fmt = "%s, %.8f, %.8f, %.2f, %.2f\n"
        with open(output_filepath, "w") as fh:
            for i in np.arange(self.n_records):
                time_str = self.time[i].isoformat()
                fh.write(fmt % (time_str, self.latitude[i], self.longitude[i], xc[i], yc[i]))

    @property
    def n_records(self):
        return len(self.time)

    @property
    def time_bounds(self):
        return [np.amin(self.time), np.amax(self.time)]

    @property
    def has_xy_coordinates(self):
        """
        Indicates if this instnace has assigned floenavi (x,y) coordinates
        :return: flag (bool)
        """
        return self.xc is not None and self.yc is not None


class WaypointData(object):
    """
    Container for individual waypoints, meant for use with dataset.Waypoints
    """

    def __init__(self, label, time, longitude, latitude, xc=None, yc=None):
        """
        Create a Waypoint entry
        :param label: (str) label of the waypoints
        :param time: (datetime.datetime) timestamp in UTC
        :param longitude: (float) longitude in degrees
        :param latitude: (float) latitude in degrees
        :param xc: (float) optional: FloeNavi x-coordinate (will be overwritten by `add_xy` method)
        :param yc: (float) optional: FloeNavi y-coordinate (will be overwritten by `add_xy` method)
        """
        self.label = label
        self.time = time
        self.longitude = longitude
        self.latitude = latitude
        self.xc = xc
        self.yc = yc

    def add_xy(self, xc, yc):
        """
        Add the FloeNavi grid coordinates to the waypoint
        :param xc: (float) FloeNavi x-coordinate
        :param yc: (float) FloeNavi x-coordinate
        :return:
        """
        self.xc = xc
        self.yc = yc
